@@ -65,7 +65,9 @@
 #define avx2 __attribute__((target("avx2,sse4.2,sse2,fpmath=sse")))
 #endif
 
-#if HAS_GCC(4, 5) && defined(__OPTIMIZE__)
+#if HAS_GCC(4, 6) && defined(__OPTIMIZE__)
+#define fast_memcpy __attribute__((optimize("Ofast"))) __attribute__((target("inline-all-stringops")))
+#elif HAS_GCC(4, 5) && defined(__OPTIMIZE__)
 #define fast_memcpy __attribute__((target("inline-all-stringops")))
 #else
 #define fast_memcpy
